/* Copyright (C) 2012 Carlos Pais 
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "wait.h"
#include "action.h"

#include <QDebug>
#include <QVariant>

static ActionEditorWidget* mEditorWidget = 0;
ActionInfo Action::Info;

Action::Action(QObject *parent) :
    QObject(parent)
{
    init();
}

Action::Action(const QVariantMap& data, QObject *parent) :
    QObject(parent)
{
    init();

    if (data.contains("object") && data.value("object").type() == QVariant::String) {
        Object* obj = ResourceManager::resource(data.value("object").toString());

        if (obj)
            mObject = obj;
    }

    if (data.contains("name") && data.value("name").type() == QVariant::String) {
        setObjectName(data.value("name").toString());
    }

    if (data.contains("skippable") && data.value("skippable").type() == QVariant::Bool) {
        mAllowSkipping = data.value("skippable").toBool();
    }

    if (data.contains("type") && data.value("type").type() == QVariant::String) {
        setType(data.value("type").toString());
    }
}

void Action::init()
{
    mObject = 0;
    mAllowSkipping = false;
    mMouseClickOnFinish = false;
    mSupportedEvents = Interaction::None;
    setType(Info.type);
    setName(Info.name);
    setIcon(Info.icon);
    connect(this, SIGNAL(dataChanged()), this, SLOT(onDataChanged()));
}


void Action::setActionEditorWidget(ActionEditorWidget* widget)
{
    mEditorWidget = widget;
}

ActionEditorWidget* Action::actionEditorWidget()
{
    return mEditorWidget;
}

ActionEditorWidget* Action::editorWidget()
{
    return mEditorWidget;
}

QIcon Action::icon() const
{
    return mIcon;
}

QString Action::description() const
{
    return mDescription;
}

bool Action::allowSkipping() const
{
    return mAllowSkipping;
}

Action* Action::newAction(QObject *parent)
{
    return new Action(parent);
}

void Action::setIcon(const QIcon & icon)
{
    mIcon = icon;
}

void Action::setDescription(const QString& desc)
{
    mDescription = desc;
}

QString Action::name() const
{
    return mName;
}

void Action::setName(const QString & name)
{
    mName = name;
}


Object* Action::sceneObject() const
{
    return mObject;
}

void Action::setSceneObject(Object * object)
{
    mObject = object;
}


void Action::onDataChanged()
{
}

void Action::paint(const QPainter & painter)
{
}

void Action::setDisplayText(const QString & text)
{
    mDisplayText = text;
    emit dataChanged();
}

QString Action::displayText() const
{
    return mDisplayText;
}

QString Action::toString() const
{
    return QString("%1 [%2]").arg(displayText()).arg(mName);
}

QString Action::type() const
{
    return mType;
}

void Action::setType(const QString & type)
{
    mType = type;
}

bool Action::supportsEvent(Interaction::InputEvent ev)
{
    return mSupportedEvents.testFlag(ev);
}

void Action::setSupportedEvents(Interaction::InputEvents events)
{
    mSupportedEvents = events;
}

bool Action::mouseClickOnFinish()
{
    return mMouseClickOnFinish;
}

void Action::setMouseClickOnFinish(bool mouseClick)
{
    mMouseClickOnFinish = mouseClick;
}

void Action::initFrom(Action* action)
{
    mMouseClickOnFinish = action->mouseClickOnFinish();
    mAllowSkipping = action->allowSkipping();
}

QVariantMap Action::toJsonObject()
{
    QVariantMap action;
    if (! objectName().isEmpty())
        action.insert("name", objectName());

    action.insert("type", mType);
    if (mAllowSkipping)
        action.insert("skippable", mAllowSkipping);
    if (mMouseClickOnFinish) {
        Wait wait;
        wait.setWaitType(Wait::MouseClick);
        action.insert("wait", wait.toJsonObject());
    }

    if (mObject)
        action.insert("object", mObject->objectName());

    return action;
}

void Action::setObjectName(const QString &name)
{
    QObject::setObjectName(name);

    emit objectNameChanged();
}

void Action::focusIn()
{
}

void Action::focusOut()
{
}
